#!/usr/bin/env python

from . import http_server
from .counter import counter
import socket, time
from rs4 import asyncore
import ssl
from skitai import lifetime
import os, sys, errno
import skitai
from errno import EWOULDBLOCK
from aquests.protocols.http2 import H2_PROTOCOLS

class https_channel (http_server.http_channel):
	ac_out_buffer_size = 65536
	ac_in_buffer_size = 65536

	def __init__(self, server, conn, addr):
		http_server.http_channel.__init__(self, server, conn, addr)

	def send(self, data):
		#print	("SEND", str (data), self.get_terminator ())
		try:
			result = self.socket.send(data)

		except ssl.SSLError as why:
			if why.errno == ssl.SSL_ERROR_WANT_WRITE:
				return 0
			elif why.errno == ssl.SSL_ERROR_ZERO_RETURN:
				self.handle_close ()
				return 0
			else:
				raise

		if result <= 0:
			return 0
		else:
			self.server.bytes_out.increment(result)
			return result

	def recv(self, buffer_size = 65535):
		try:
			result = self.socket.recv(buffer_size)
			#print ("~~~~~~~~~~~~~", len (result), result)
			if result is None:
				return b''

			elif result == b'':
				self.handle_close()
				return b''

			else:
				self.server.bytes_in.increment(len(result))
				return result

		except MemoryError:
			lifetime.shutdown (1, 1.0)

		except ssl.SSLError as why:
			if why.errno == ssl.SSL_ERROR_WANT_READ:
				try:
					raise BlockingIOError
				except NameError:
					raise socket.error (EWOULDBLOCK)
			# closed connection
			elif why.errno in (ssl.SSL_ERROR_ZERO_RETURN, ssl.SSL_ERROR_EOF):
				self.handle_close ()
				return b''
			else:
				raise


class https_server (http_server.http_server):
	def __init__ (self, ip, port, ctx, server_logger = None, request_logger = None):
		http_server.http_server.__init__ (self, ip, port, server_logger, request_logger)
		self.ctx = ctx
		self.socket = self.ctx.wrap_socket (self.socket, server_side = True)

	def handle_accept (self):
		self.total_clients.inc()

		try:
			conn, addr = self.accept()
		except socket.error:
			#self.log_info ('server accept() threw an exception', 'warning')
			return
		except TypeError:
			if os.name == "nt":
				self.log_info ('server accept() threw EWOULDBLOCK', 'warning')
			return
		except:
			self.trace()

		https_channel (self, conn, addr)


def init_context (certfile, keyfile, pass_phrase):
	try:
		protocol = ssl.PROTOCOL_TLS
	except AttributeError:
		protocol = ssl.PROTOCOL_SSLv23
	ctx = ssl.SSLContext (protocol)
	try:
		ctx.set_alpn_protocols (H2_PROTOCOLS)
	except AttributeError:
		ctx.set_npn_protocols (H2_PROTOCOLS)
	ctx.load_cert_chain (certfile, keyfile, pass_phrase)
	ctx.check_hostname = False
	return ctx


if __name__ == "__main__":
	import module_loader
	from aquests.athreads import threadlib
	import file_handler, xmlrpc_handler, soap_handler, cgi_handler, graph_handler, proxy_handler, logger

	pools = threadlib.request_queue()
	daemons = {}
	for i in range (2):
		d=threadlib.request_thread (pools, 8)
		daemons [i+1]=d
		d.start()

	server_logger = framework.ServerLogger('%slog2/' % os.environ ['ALEPH_HOME'])
	modules = module_loader.ModuleLoader('%smod/' % os.environ ['ALEPH_HOME'], ['hello', 'iserver', 'system'], logger = server_logger)
	ctx = init_context('sslv3', 'cert/server.csr', 'cert/ca.crt', 'fatalbug')
	sv = https_server ('', 9443, ctx, https_channel, server_logger)

	class ServerComponent:
		queue = pools
		server = sv
		modules = modules
		channels = asyncore.socket_map
		daemons = daemons
	server=ServerComponent

	sh=soap_handler.soap_handler(server, 'http://infogent.sseki.com')
	sv.install_handler(sh)

	xh=xmlrpc_handler.xmlrpc_handler(server)
	sv.install_handler(xh)

	ch=cgi_handler.cgi_handler(server)
	sv.install_handler(ch)

	gh=graph_handler.graph_handler()
	sv.install_handler(gh)

	ph=proxy_handler.proxy_handler(server)
	sv.install_handler(ph)

	fh=file_handler.file_handler('/home/infogent/infogent/doc/')
	sv.install_handler(fh)

	if os.name == 'posix': usepoll = 1
	else: usepoll = 0

	Rand.load_file('randpool.dat', -1)
	try:
		asyncore.loop(30.0, usepoll)
	except:
		server_logger.trace()
	Rand.save_file('randpool.dat')
	os.abort ()



#!/usr/bin/env python

import sys
from rs4 import asyncore, asynchat
import re, socket, time, threading, os
from . import http_request, counter
from aquests.protocols.http import http_util, http_date
from aquests.athreads import threadlib
from skitai import lifetime
from rs4 import producers, compressors
from collections import deque
import signal
import ssl
import skitai
from hashlib import md5
from rs4.psutil.processutil import set_process_name
if os.name == "posix":
	import psutil
	CPUs = psutil.cpu_count()

PID = {}
ACTIVE_WORKERS = 0
SURVAIL = True
EXITCODE = 0
DEBUG = False


#-------------------------------------------------------------------
# server channel
#-------------------------------------------------------------------
class http_channel (asynchat.async_chat):
	current_request = None
	channel_count = counter.counter ()
	closed = False
	is_rejected = False

	keep_alive = 30
	network_timeout = 30
	zombie_timeout = 30

	fifo_class = deque

	def __init__ (self, server, conn, addr):
		self.channel_number = http_channel.channel_count.inc ()
		self.request_counter = counter.counter()
		self.bytes_out = counter.counter()
		self.bytes_in = counter.counter()

		#asynchat.async_chat.__init__ (self, conn)
		self.ac_in_buffer = b''
		self.incoming = []
		self.producer_fifo = self.fifo_class ()
		asyncore.dispatcher.__init__(self, conn)

		self.server = server
		self.addr = addr
		self.set_terminator (b'\r\n\r\n')
		self.in_buffer = b''
		self.creation_time = int (time.time())
		self.event_time = int (time.time())
		self.__history = []
		self.__sendlock = None
		self.producers_attend_to = []
		self.things_die_with = []

	def use_sendlock (self):
		self.__sendlock  = threading.Lock ()
		self.initiate_send = self._initiate_send_ts

	def get_history (self):
		return self.__history

	def reject (self):
		self.is_rejected = True

	def _initiate_send_ts (self):
		lock = self.__sendlock
		lock.acquire ()
		try:
			asynchat.async_chat.initiate_send (self)
			try: is_working = self.producer_fifo.working ()
			except AttributeError: 	is_working = len (self.producer_fifo)
		finally:
			lock.release ()
		if not is_working:
			self.done_request ()

	def initiate_send (self):
		asynchat.async_chat.initiate_send (self)
		try: is_working = self.producer_fifo.working ()
		except AttributeError: 	is_working = len (self.producer_fifo)
		if not is_working:
			self.done_request ()

	def readable (self):
		return not self.is_rejected and asynchat.async_chat.readable (self)

	def issent (self):
		return self.bytes_out.as_long ()

	def __repr__ (self):
		ar = asynchat.async_chat.__repr__(self) [1:-1]
		return '<%s channel#: %s requests:%s>' % (
				ar,
				self.channel_number,
				self.request_counter
				)

	def clean_shutdown_control (self, phase, time_in_this_phase):
		if phase == 3:
			self.reject ()
			if self.writable ():
				return 1
			self.close ()
			return 0

	def isconnected (self):
		return self.connected

	def handle_timeout (self):
		self.log ("killing zombie channel %s" % ":".join (map (str, self.addr)))
		self.close ()

	def set_timeout (self, timeout):
		self.zombie_timeout = timeout

	def set_socket_timeout (self, timeout):
		self.keep_alive = timeout
		self.network_timeout = timeout

	def attend_to (self, thing):
		if not thing: return
		self.producers_attend_to.append (thing)

	def die_with (self, thing, tag):
		if not thing: return
		self.things_die_with.append ((thing, tag))

	def done_request (self):
		self.producers_attend_to = [] # all producers are finished
		self.set_timeout (self.keep_alive)

	def send (self, data):
		#print	("SEND", repr (data), self.get_terminator ())
		self.event_time = int (time.time())
		result = asynchat.async_chat.send (self, data)
		self.server.bytes_out.inc (result)
		self.bytes_out.inc (result)
		return result

	def recv (self, buffer_size):
		self.event_time = int (time.time())
		try:
			result = asynchat.async_chat.recv (self, buffer_size)
			lr = len (result)
			self.server.bytes_in.inc (lr)
			self.bytes_in.inc (lr)
			if not result:
				self.handle_close ()
				return b""
			#print	("RECV", repr(result), self.get_terminator ())
			return result

		except MemoryError:
			lifetime.shutdown (1, 1.0)

	def collect_incoming_data (self, data):
		#print ("collect_incoming_data", repr (data [:180]), self.current_request)
		if self.current_request:
			self.current_request.collect_incoming_data (data)
		else:
			self.in_buffer = self.in_buffer + data

	def forward_domain (self, r):
		host = r.get_header ('host', '').split (":", 1)[0]
		if host == self.server.single_domain:
			return False
		scheme, port = r.get_scheme (), self.server.port
		if (scheme == 'https' and port == 443) or (scheme == 'http' and port == 80):
			port = ""
		else:
			port = ":{}".format (port)
		newloc = '{}://{}{}{}'.format (scheme, self.server.single_domain, port, r.uri)
		r.response.set_header ("Location", newloc)
		r.response.error (301, why = "Object moved to <a href='{}'>here</a>".format (newloc))
		return True

	def found_terminator (self):
		if self.is_rejected:
			return

		if self.current_request:
			self.current_request.found_terminator()
			if DEBUG:
				self.__history.append ("REQUEST DONE")
				self.__history = self.__history [-30:]

		else:
			header = self.in_buffer
			self.in_buffer = b''
			lines = header.decode ("utf8").split('\r\n')
			while lines and not lines[0]:
				lines = lines[1:]

			if not lines:
				self.close_when_done()
				return

			request = lines[0]
			try:
				command, uri, version = http_util.crack_request (request)
			except:
				self.log_info ("channel-%s invaild request header" % self.channel_number, "fail")
				return self.close ()

			if DEBUG:
				self.__history.append ("START REQUEST: %s/%s %s" % (command, version, uri))

			header = http_util.join_headers (lines[1:])
			r = http_request.http_request (self, request, command, uri, version, header)
			if self.server.single_domain and self.forward_domain (r):
				return
			self.set_timeout (self.network_timeout)
			self.request_counter.inc()
			self.server.total_requests.inc()

			if command is None:
				r.response.error (400)
				return

			for h in self.server.handlers:
				if h.match (r):
					try:
						self.current_request = r
						h.handle_request (r)

					except:
						self.server.trace()
						try: r.response.error (500)
						except: pass

					return

			try: r.response.error (404)
			except: pass

	def handle_abort (self):
		self.close (ignore_die_partner = True)
		self.log_info ("channel-%s aborted" % self.channel_number, "info")

	def close (self, ignore_die_partner = False):
		if self.closed:
			return

		for closable, tag in self.things_die_with:
			if closable and hasattr (closable, "channel"):
				closable.channel = None
			self.journal (tag)
			if not ignore_die_partner:
				self.producers_attend_to.append (closable)

		if self.current_request is not None:
			self.producers_attend_to.append (self.current_request.collector)
			self.producers_attend_to.append (self.current_request.producer)
			# 1. close forcely or by error, make sure that channel is None
			# 2. by close_when_done ()
			self.current_request.channel = None
			self.current_request = None

		for closable in self.producers_attend_to:
			if closable and hasattr (closable, "close"):
				try:
					closable.close ()
				except:
					self.server.trace()

		self.producers_attend_to, self.things_die_with = [], []
		self.discard_buffers ()
		asynchat.async_chat.close (self)
		self.connected = False
		self.closed = True
		self.log_info ("channel-%s closed" % self.channel_number, "info")

	def journal (self, reporter):
		self.log (
			"%s closed, client %s:%s, bytes in: %s, bytes out: %s for %d seconds " % (
				reporter,
				self.addr [0],
				self.addr [1],
				self.bytes_in,
				self.bytes_out,
				time.time () - self.creation_time
			)
		)

	def log (self, message, type = "info"):
		self.server.log (message, type)
	log_info = log

	def trace (self, id = None):
		self.server.trace (id)

	def handle_expt(self):
		self.log_info ("channel-%s panic" % self.channel_number, "fail")
		self.close ()

	def handle_error (self):
		self.server.trace ("channel-%s" % self.channel_number)
		self.close()


#-------------------------------------------------------------------
# server class
#-------------------------------------------------------------------
class http_server (asyncore.dispatcher):
	SERVER_IDENT = skitai.NAME

	maintern_interval = 0
	critical_point_cpu_overload = 90.0
	critical_point_continuous = 3

	def __init__ (self, ip, port, server_logger = None, request_logger = None):
		global PID

		self.handlers = []
		self.ip = ip
		self.port = port
		asyncore.dispatcher.__init__ (self)

		if ip.find (":") != -1:
			self.create_socket (socket.AF_INET6, socket.SOCK_STREAM)
		else:
			self.create_socket (socket.AF_INET, socket.SOCK_STREAM)

		self.set_reuse_addr ()
		try:
			self.bind ((ip, port))
		except OSError as why:
			if why.errno == 98:
				server_logger ("address already in use, cannot start server", "fatal")
			else:
				server_logger.trace ()
			sys.exit (0)

		self.worker_ident = "master"
		self.server_logger = server_logger
		self.request_logger = request_logger
		self.start_time = time.ctime(time.time())
		self.start_time_int = time.time()

		self.total_clients = counter.counter()
		self.total_requests = counter.counter()
		self.exceptions = counter.counter()
		self.bytes_out = counter.counter()
		self.bytes_in  = counter.counter()
		self.shutdown_phase = 2

		self.__last_maintern = time.time ()

		host, port = self.socket.getsockname()
		hostname = socket.gethostname()
		try:
			self.server_name = socket.gethostbyaddr (hostname)[0]
		except socket.error:
			self.server_name = hostname
		self.hash_id = md5 (self.server_name.encode ('utf8')).hexdigest() [:4]
		self.server_port = port
		self.single_domain = None

	def set_single_domain (self, domain):
		self.single_domain = domain

	def _serve (self, shutdown_phase = 2):
		self.shutdown_phase = shutdown_phase
		self.listen (os.name == "posix" and 4096 or 256)

	def serve (self, sub_server = None):
		if sub_server:
			sub_server._serve (max (1, self.shutdown_phase - 1))
		self._serve ()

	def fork (self, numworker = 1):
		global ACTIVE_WORKERS, SURVAIL, PID, EXITCODE

		if os.name == "nt":
			set_process_name (skitai.get_proc_title ())
			signal.signal(signal.SIGTERM, hTERMWORKER)

		else:
			while SURVAIL:
				try:
					if ACTIVE_WORKERS < numworker:
						pid = os.fork ()
						if pid == 0:
							self.worker_ident = "w%d" % len (PID)
							set_process_name ("%s:%s" % (skitai.get_proc_title (), self.worker_ident))
							PID = {}
							signal.signal(signal.SIGTERM, hTERMWORKER)
							signal.signal(signal.SIGQUIT, hQUITWORKER)
							break

						else:
							set_process_name ("%s:m" % skitai.get_proc_title ())
							if not PID:
								signal.signal(signal.SIGHUP, hHUPMASTER)
								signal.signal(signal.SIGTERM, hTERMMASTER)
								signal.signal(signal.SIGINT, hTERMMASTER)
								signal.signal(signal.SIGQUIT, hQUITMASTER)
								signal.signal (signal.SIGCHLD, hCHLD)

							ps = psutil.Process (pid)
							ps.x_overloads = 0
							PID [pid] = ps
							ACTIVE_WORKERS += 1
							#print ('-----', PID, ACTIVE_WORKERS)

					now = time.time ()
					if self.maintern_interval and now - self.__last_maintern > self.maintern_interval:
						self.maintern (now)
					time.sleep (1)

				except KeyboardInterrupt:
					EXITCODE = 0
					DO_SHUTDOWN (signal.SIGTERM)
					break

			if self.worker_ident == "master":
				return EXITCODE

		self.log_info ('%s (%s) started on %s:%d' % (
			self.SERVER_IDENT, self.worker_ident, self.server_name, self.port)
		)

	usages = []
	cpu_stat = []
	def maintern (self, now):
		global PID, CPUs

		self.__last_maintern = now
		usages = []
		for ps in PID.values ():
			if ps is None:
				continue

			try:
				usage = ps.cpu_percent ()
			except (psutil.NoSuchProcess, AttributeError):
				# process changed, maybe next time
				usage = 0.0
			usages.append ((ps, usage))

		if not usages:
			# restarting?
			return

		self.usages.append (sum ([x [1] for x in usages]) / len (usages))
		self.usages = self.usages [-45:]

		self.cpu_stat = []
		max_usage = 0
		for c in (3, 15, 45):
			# 1m, 5m, 15
			array = self.usages [-c:]
			self.cpu_stat.append ((max (array), int (sum (array) / c)))

		if self.cpu_stat [0][0] > 10:
			# 60% for 1m
			self.log ("CPU percent: " + ", ".join (["%d/%d" % unit for unit in self.cpu_stat]))

		for ps, usage in usages:
			if usage < self.critical_point_cpu_overload:
				ps.x_overloads = 0 #reset count
				continue

			ps.x_overloads += 1
			if ps.x_overloads > self.critical_point_continuous:
				self.log ("process %d is overloading, try to kill..." % ps.pid, 'fatal')
				sig = ps.x_overloads > (self.critical_point_continuous + 2) and signal.SIGKILL or signal.SIGTERM
				try:
					os.kill (ps.pid, sig)
				except OSError:
					pass

	def create_socket(self, family, type):
		if hasattr (socket, "_no_timeoutsocket"):
			sock_class = socket._no_timeoutsocket
		else:
			sock_class = socket.socket

		self.family_and_type = family, type
		sock = sock_class (family, type)
		sock.setblocking(0)
		self.set_socket(sock)

	def clean_shutdown_control (self, phase, time_in_this_phase):
		if phase == self.shutdown_phase:
			self.log_info ('shutting down web server: %s' % self.server_name)
			self.close ()

	def close (self):
		asyncore.dispatcher.close (self)
		for h in self.handlers:
			if hasattr (h, "close"):
				h.close ()

	def writable (self):
		return 0

	def install_handler (self, handler, back = 1):
		if back:
			self.handlers.append (handler)
		else:
			self.handlers.insert (0, handler)

	def remove_handler (self, handler):
		self.handlers.remove (handler)

	def log (self, message, type = "info"):
		if self.server_logger:
			self.server_logger.log (message, type)
		else:
			sys.stdout.write ('log: [%s] %s\n' % (type,str (message)))

	def log_request (self, message):
		if self.request_logger:
			self.request_logger.log (message, "")
		else:
			sys.stdout.write ('%s\n' % message)

	def log_info(self, message, type='info'):
		self.log (message, type)

	def trace (self, id = None):
		self.exceptions.inc()
		if self.server_logger:
			self.server_logger.trace (id)
		else:
			asyncore.dispatcher.handle_error (self)

	def handle_read (self):
		pass

	def readable (self):
		return self.accepting

	def handle_error (self):
		self.trace()

	def handle_connect (self):
		pass

	def handle_accept (self):
		#print ('-----', self.worker_ident)
		self.total_clients.inc()
		try:
			conn, addr = self.accept()
		except socket.error:
			self.log_info ('server accept() threw an exception', 'warn')
			return
		except TypeError:
			if os.name == "nt":
				self.log_info ('server accept() threw EWOULDBLOCK', 'warn')
			return
		#self.log_info ('client %s:%d accepted by %s' % (addr [0], addr [1], self.worker_ident))
		http_channel (self, conn, addr)

	def handle_expt (self):
		self.log_info ('socket panic', 'warning')

	def handle_close (self):
		self.log_info('server shutdown', 'warning')
		self.close()

	def status (self):
		global PID

		return 	{
			"child_pids": list (PID.keys ()),
			"ident": "%s for %s" % (self.worker_ident, self.SERVER_IDENT),
			'server_name': self.server_name,
			"start_time": self.start_time,
			'hash_id': self.hash_id,
			"port": self.port,
			"total_clients": self.total_clients.as_long(),
			"total_request": self.total_requests.as_long(),
			"total_exceptions": self.exceptions.as_long(),
			"bytes_out": self.bytes_out.as_long(),
			"bytes_in": self.bytes_in.as_long(),
			"cpu_percent":	self.cpu_stat
		}

def hCHLD (signum, frame):
	global ACTIVE_WORKERS, PID

	ACTIVE_WORKERS -= 1
	try:
		pid, status = os.waitpid (0, 0)
	except ChildProcessError:
		pass
	else:
		PID [pid]	= None

def hTERMWORKER (signum, frame):
	lifetime.shutdown (0, 1.0)

def hQUITWORKER (signum, frame):
	lifetime.shutdown (0, 30.0)

def DO_SHUTDOWN (sig):
	global PID

	signal.signal (signal.SIGCHLD, signal.SIG_IGN)
	for pid in PID:
		if PID [pid] is None:
			continue
		try: os.kill (pid, sig)
		except OSError: pass

def hTERMMASTER (signum, frame):
	global EXITCODE, SURVAIL
	SURVAIL = False
	EXITCODE = 0
	DO_SHUTDOWN (signal.SIGTERM)

def hQUITMASTER (signum, frame):
	global EXITCODE
	EXITCODE = 0
	DO_SHUTDOWN (signal.SIGQUIT)

def hHUPMASTER (signum, frame):
	global EXITCODE
	EXITCODE = 3
	DO_SHUTDOWN (signal.SIGTERM)

def configure (name, network_timeout = 0, keep_alive = 0):
	from . import https_server
	http_server.SERVER_IDENT = name
	https_server.https_server.SERVER_IDENT = name + " (SSL)"
	http_channel.keep_alive = https_server.https_channel.keep_alive = not keep_alive and 2 or keep_alive
	http_channel.network_timeout = https_server.https_channel.network_timeout = not network_timeout and 30 or network_timeout


if __name__ == "__main__":
	pass



import sys, os
try:
	from urllib.parse import unquote
except ImportError:
	from urllib import unquote
import sys
from rs4 import producers
from skitai.http_response import catch
from aquests.athreads import trigger
from . import collectors
from skitai import version_info, was as the_was
import threading
try:
	from cStringIO import StringIO as BytesIO
except ImportError:
	from io import BytesIO
import skitai
from ..utility import make_pushables

header2env = {
	'content-length'	: 'CONTENT_LENGTH',
	'content-type'	  : 'CONTENT_TYPE',
	'connection'		: 'CONNECTION_TYPE'
	}

SKITAI_VERSION = ".".join (map (lambda x: str (x), version_info [:3]))

class Handler:
	GATEWAY_INTERFACE = 'CGI/1.1'
	ENV = {
			'GATEWAY_INTERFACE': 'CGI/1.1',
			'SERVER_SOFTWARE': "Skitai App Engine/%s Python/%d.%d" % ((SKITAI_VERSION,) + sys.version_info[:2]),
			'skitai.version': tuple (version_info [:3]),
			"wsgi.version": (1, 0),
			"wsgi.errors": sys.stderr,
			"wsgi.run_once": False,
			"wsgi.input": None
	}

	def __init__(self, wasc, apps = None):
		self.wasc = wasc
		self.apps = apps
		self.ENV ["skitai.process"] = self.wasc.workers
		self.ENV ["skitai.thread"] = 0
		if hasattr (self.wasc, "threads") and self.wasc.threads:
			self.ENV ["skitai.thread"] = len (self.wasc.threads)
		self.ENV ["wsgi.multithread"] = hasattr (self.wasc, "threads") and self.wasc.threads
		self.ENV ["wsgi.url_scheme"] = hasattr (self.wasc.httpserver, "ctx") and "https" or "http"
		self.ENV ["wsgi.multiprocess"] = self.wasc.workers > 1 and os.name != "nt"
		self.ENV ['SERVER_PORT'] = str (self.wasc.httpserver.port)
		self.ENV ['SERVER_NAME'] = self.wasc.httpserver.server_name

	def match (self, request):
		return 1

	def build_environ (self, request, apph):
		(path, params, query, fragment) = request.split_uri()
		if params: path = path + params
		while path and path[0] == '/':
			path = path[1:]
		if '%' in path: path = unquote (path)
		if query: query = query[1:]

		env = self.ENV.copy ()
		env ['REQUEST_METHOD'] = request.command.upper()
		env ['REQUEST_URI'] = request.uri
		env ['REQUEST_VERSION'] = request.version
		env ['SERVER_PROTOCOL'] = "HTTP/" + request.version
		env ['CHANNEL_CREATED'] = request.channel.creation_time
		if query: env['QUERY_STRING'] = query
		env ['REMOTE_ADDR'] = request.channel.addr [0]
		env ['REMOTE_SERVER'] = request.channel.addr [0]
		env ['SCRIPT_NAME'] = apph.route
		env ['SCRPIT_PATH'] = apph.abspath
		env ['PATH_INFO'] = apph.get_path_info (path)

		for header in request.header:
			key, value = header.split(":", 1)
			key = key.lower()
			value = value.strip ()
			if key in header2env and value:
				env [header2env [key]] = value
			else:
				key = 'HTTP_%s' % ("_".join (key.split ( "-"))).upper()
				if value and key not in env:
					env [key] = value

		for k, v in list(os.environ.items ()):
			if k not in env:
				env [k] = v

		return env

	def make_collector (self, collector_class, request, max_cl, *args, **kargs):
		collector = collector_class (self, request, *args, **kargs)

		if collector.content_length is None:
			if not request.version.startswith ("2."):
				del collector
				self.handle_error_before_collecting (request, 411)
				return
			else:
				collector.set_max_content_length (max_cl)

		elif collector.content_length > max_cl: #5M
			self.wasc.logger ("server", "too large request body (%d bytes)" % collector.content_length, "wran")
			del collector
			if request.get_header ("expect") == "100-continue":
				self.handle_error_before_collecting (request, 413, False) # client doesn't send data any more, I wish.
			else:
				self.handle_error_before_collecting (request, 413)	# forcely disconnect
			return

		# ok. allow form-data
		if request.get_header ("expect") == "100-continue":
			request.response.instant ("100 Continue")

		return collector

	def handle_error_before_collecting (self, request, code, force_close = True):
		if request.version != "2.0":
			if request.command in ('post', 'put', 'patch') and force_close:
				request.response.abort (code)
			else:
				request.response.error (code)
		else:
			# keep connecting on HTTP/2 as possible
			if request.command in ('post', 'put', 'patch'):
				collector = collectors.HTTP2DummyCollector (self, request, code)
				request.collector = collector
				collector.start_collect ()
			else:
				request.response.error (code)

	def handle_request (self, request):
		path, params, query, fragment = request.split_uri ()
		has_route = self.apps.has_route (path)
		if has_route == 0:
			return self.handle_error_before_collecting (request, 404)
		if has_route == 1:
			request.response ["Location"] = "%s/" % path
			return self.handle_error_before_collecting (request, 308)

		app = self.apps.get_app (has_route).get_callable()
		# for rendering error template
		request.response.current_app = app

		if request.command != "options" and skitai.HAS_ATILA and isinstance (app, skitai.HAS_ATILA):
			# pass through options, because options want authentification info.
			if not app.is_authorized (request, app.authenticate):
				self.handle_error_before_collecting (request, 401)
				return
			if not app.is_allowed_origin (request, app.access_control_allow_origin):
				self.handle_error_before_collecting (request, 403)
				return

		if request.command in ('post', 'put', 'patch'):
			try:
				# shoud have constructor __init__ (self, handler, request, upload_max_size, file_max_size, cache_max_size)
				collector_class = app.get_collector (request)
			except AttributeError:
				collector_class = None
			except NotImplementedError:
				# proto buf is not ready
				return self.handle_error_before_collecting (request, 501)

			ct = request.get_header ("content-type", "")
			if ct.startswith ("multipart/form-data"):
				max_size = app.config.max_multipart_body_size
			else:
				max_size = app.config.max_post_body_size

			if collector_class is None:
				if ct.startswith ("multipart/form-data"):
					collector_class = collectors.MultipartCollector
				else:
					collector_class = collectors.FormCollector

			args = (
				app.config.max_multipart_body_size,
				app.config.max_upload_file_size,
				app.config.max_cache_size
			)
			collector = self.make_collector (collector_class, request, max_size, *args)
			if collector:
				request.collector = collector
				collector.start_collect ()

		elif request.command in ('get', 'delete', 'options'):
			self.continue_request(request)

		else:
			self.handle_error_before_collecting (request, 405)

	def continue_request (self, request, data = None, respcode = None):
		if respcode: # delayed resp code on POST
			if respcode [1]:
				# force close
				return request.response.abort (respcode [0])
			return request.response.error (respcode [0])

		try:
			path, params, query, fragment = request.split_uri ()
			apph = self.apps.get_app (path)

		except:
			self.wasc.logger.trace ("server",  request.uri)
			return request.response.error (500, why = apph.debug and sys.exc_info () or None)

		try:
			env = self.build_environ (request, apph)
			if data:
				env ["wsgi.input"] = data
			args = (env, request.response.start_response)

		except:
			self.wasc.logger.trace ("server",  request.uri)
			return request.response.error (500, why = apph.debug and csys.exc_info () or None)

		if env ["wsgi.multithread"]:
			self.wasc.queue.put (Job (request, apph, args, self.apps, self.wasc.logger))
		else:
			Job (request, apph, args, self.apps, self.wasc.logger) ()


class Job:
	# Multi-Threaded Jobs
	def __init__(self, request, apph, args, apps, logger):
		self.request = request
		self.apps = apps
		self.apph = apph
		self.args = args
		self.logger = logger

	def __repr__(self):
		return "<Job %s %s HTTP/%s>" % (self.request.command.upper (), self.request.uri, self.request.version)

	def __str__ (self):
		return "%s %s HTTP/%s" % (self.request.command.upper (), self.request.uri, self.request.version)

	def exec_app (self):
		# this is not just for atila,
		# Corequest need request and response
		was = the_was._get ()
		was.apps = self.apps
		was.request = self.request
		was.response = self.request.response
		was.env = self.args [0]
		was.env ["skitai.was"] = was

		response = self.request.response
		try:
			content = self.apph (*self.args)
			will_be_push = make_pushables (response, content)
		except MemoryError:
			raise
		except:
			was.traceback ()
			trigger.wakeup (lambda p = response, d=self.apph.debug and sys.exc_info () or None: (p.error (500, "Internal Server Error", d), p.done ()) )
		else:
			if will_be_push is None: # not responsible or futures
				return
			for part in will_be_push:
				if len (will_be_push) == 1 and type (part) is bytes and len (response) == 0:
					response.update ("Content-Length", len (part))
				response.push (part)
			trigger.wakeup (lambda p = response: (p.done (),))

	def __call__(self):
		try:
			try:
				self.exec_app ()
			finally:
				self.deallocate	()
		except:
			# no response, alredy done. just log
			self.logger.trace ("server",  self.request.uri)

	def deallocate (self):
		env = self.args [0]
		_input = env ["wsgi.input"]
		if _input:
			try: _input.close ()
			except AttributeError: pass
			if hasattr (_input, "name"):
				try: os.remove (_input.name)
				except: self.logger.trace ("app")

		was = env.get ("skitai.was")
		if was is not None:
			was.apps = None
			was.env = None
			try: del was.response
			except AttributeError: pass
			try: del was.request
			except AttributeError: pass



from skitai import counter
from rs4 import compressors
from hashlib import md5
from rs4 import pathtool
import os
import time
import threading
import glob
import threading
from functools import reduce

class CacheInfo:
	def __init__ (self):		
		self.count = 0
		self.allocated = 0.0
		self.accumulated_caches = 0
		self.accumulated_expires = 0
		
	def inc (self, delta):
		self.count += 1
		self.allocated += delta
		self.accumulated_caches += 1
		
	def dec (self, delta):
		self.count -= 1
		self.allocated -= delta
		self.accumulated_expires += 1
	
	def get (self):
		return (
			self.count, 
			self.allocated, 
			self.accumulated_caches, 
			self.accumulated_expires
		)	
		
		
class CacheFileSystem:
	maintern_interval = 600
	
	def __init__ (self, path = None, memmax = 64, diskmax = 0):
		self.path = path
		if not self.path:
			diskmax = 0
		self.max_memory = memmax * 1024 * 1024
		self.max_disk = diskmax * 1024 *1024		
		self.files = {}
		self.current_memory = CacheInfo ()
		self.current_disk = CacheInfo ()
		self.maintern_times = {}
		self.lock = threading.RLock ()		
		self.mainterns = counter.counter ()
		self.numhits = counter.counter ()
		self.numfails = counter.counter ()
		if self.max_disk:
			self.check_dir ()
	
	def status (self):
		with self.lock:
			r = {
				"numhits": self.numhits.as_long (),
				"numfails": self.numfails.as_long (),	
				'mainterned':self.mainterns.as_long (),
				"memory": self.current_memory.get (),
				"disk": self.current_disk.get ()
			}
		return r

	def check_dir (self):
		for h in list(range(48,58)) + list(range(97,103)):
			for i in list(range(48,58)) + list(range(97,103)):
				for j in list(range(48,58)) + list(range(97,103)):
					initial = "0" + chr (h) + "/" + chr (i) + chr (j)
					pathtool.mkdir (os.path.join (self.path, initial))
					self.load_cache (initial)
	
	def load_cache (self, initial, force_remove = False):
		c = 0
		current_time = time.time ()
		for path in glob.glob (os.path.join (self.path, initial, "*")):
			created = os.stat (path).st_mtime
			
			try: self.files [initial]
			except KeyError: 	self.files [initial] = {}
			
			fn = os.path.split (path)[-1]
			f = open (path, "rb")
			max_age = int (f.read (12).strip ())
			compressed = int (f.read (1).strip ())
			f.close ()
			
			if created < current_time - max_age:				
				try: os.remove (path)				
				except (OSError, IOError): pass
				continue
				
			size = os.path.getsize (path)
			self.files [initial][fn] = (created, 1, size, compressed, max_age)
			with self.lock:
				self.current_disk.inc (size)
					
	def maintern (self, initial, force_remove = False):
		current_time = time.time ()		
		deletables = []	
		for fn in self.files [initial]:
			cached = self.files [initial][fn]
			valid_time = current_time - cached [4] # max-age
			if cached [0] < valid_time or force_remove:
				deletables.append ((fn, cached [1], cached [2]))
		
		for fn, memhit, usage in deletables:		
			if memhit == 1:
				with self.lock:			
					self.current_memory.dec (usage)					
				del self.files [initial][fn]				
			elif self.path:	
				try: os.remove (os.path.join (self.path, initial, fn))
				except (OSError, IOError): pass
				else:	
					with self.lock:			
						self.current_disk.dec (usage)
					del self.files [initial][fn]				
		
		self.maintern_times [initial] = current_time
		with self.lock:
			self.mainterns.inc ()
			
	def truncate (self):
		for initial in self.files.keys ():
			self.maintern (initial, True)
						
	def getpath (self, key):
		key = key.encode ("utf8")		
		file = md5 (key).hexdigest ()
		initial = "0" + file [0] + "/" + file [1:3]
		if self.path:	
			path = os.path.join (self.path, initial, file), initial, file
		else:
			path = None			
		return path, initial, file	
	
	def is_cachable (self, cache_control, has_cookie, has_auth, pragma):
		if pragma == "no-cache":			
			return False		
		if cache_control:			
			cache_control = list (map (lambda x: x.strip (), cache_control.split (",")))
			if "no-cache" in cache_control or "max-age=0" in cache_control:
				return False
			if has_auth and "public" not in cache_control:
				cachable = False
			if has_cookie and "public" not in cache_control:
				cachable = False						
		return True
					
	def get (self, key, undecompressible = 0):			
		path, initial, fn = self.getpath (key)
		
		try:
			cached = self.files [initial][fn]
		except KeyError:
			with self.lock:
				self.numfails.inc ()
			return None, None, None, None, None
		
		current_time = int (time.time ())
		memhit = cached [1]
		max_age = cached [4]
				
		if cached [0] < current_time - max_age:
			if memhit == 1:
				del self.files [initial][fn]
				with self.lock:
					self.current_memory.dec (cached [2])
			elif path:
				try: 
					os.remove (path)
				except (OSError, IOError): 
					pass
				else:
					del self.files [initial][fn]
					with self.lock:
						self.current_disk.dec (cached [2])
			return None, None, None, None, None
		
		with self.lock:
			self.numhits.inc ()
		
		if memhit == 1:
			content_type, content = cached [-2:]			
		
		else:	
			f = open (path, "rb")
			f.read (13) # abandon max_age, compressed
			content_type = f.read (64).strip ()
			content = f.read ()
			f.close ()			
		
		compressed = cached [3]
		if compressed and not undecompressible:
			decompressor = compressors.GZipDecompressor ()
			content = decompressor.decompress (content) + decompressor.flush ()
			compressed = 0
		
		return memhit, compressed, max_age, content_type, content
	
	def save (self, key, content_type, content, max_age, compressed = 0):		
		if self.max_memory == 0 or not self.max_disk == 0:
			return
			
		usage = len (content)
		if usage > 10000000:
			return
		
		if len (str (max_age)) > 12 or len (content_type) > 64:
			return
		
		# check memory status
		with self.lock:
			current_memory = self.current_memory.get ()[1]		
			current_disk = self.current_disk.get ()[1]
		if current_memory > self.max_memory and current_disk > self.max_disk:
			# there's no memory/disk room for cache
			return self.maintern (initial)
		
		current_time = int (time.time ())
		path, initial, fn = self.getpath (key)
		try: self.files [initial]
		except KeyError: self.files [initial] = {}
		else:
			with self.lock:
				last_maintern = self.maintern_times.get (initial, 0.)					
			if last_maintern == 0:
				self.maintern_times [initial] = current_time
			elif current_time - last_maintern > self.maintern_interval:
				self.maintern (initial)	
		
		# already have valid cache
		cached = self.files [initial].get (fn)
		if cached: # already have
			return
			
		if not compressed and content_type.startswith ("text/"):
			compressor = compressors.GZipCompressor ()
			content = compressor.compress (content) + compressor.flush ()
			compressed = 1
			usage = len (content)
		
		if current_memory <= self.max_memory:
			usage *= 1.5
			self.files [initial][fn] = (current_time, 1, usage, compressed, max_age, content_type, content)
			with self.lock:
				self.current_memory.inc (usage)
			return
		
		if self.max_disk:
			f = open (path, "wb")
			if not content_type:
				content_type = b""			
			f.write (("%12s%d%64s" % (max_age, compressed, content_type)).encode ("utf8"))
			f.write (content)
			f.close ()
			self.files [initial][fn] = (current_time, -1, usage, compressed, max_age)
			with self.lock:
				self.current_disk.inc (usage)

			
if __name__ == "__main__":
	f = ReverseProxy ("g:\\ttt")
	f.save ("a", "b", "asdasda")
	print(f.get ("a", "b"))
	print(f.status ())



import threading
from . import specs
try: from urllib.parse import quote_plus
except ImportError: from urllib import quote_plus
import skitai

class WebSocketServers:
	def __init__ (self, wasc):
		self.wasc = wasc
		self._nomore = False
		self.lock = threading.RLock ()
		self.wss = {}

	def get (self, gid, default = None):
		with self.lock:
			return self.wss.get (gid, default)

	def has_key (self, gid):
		with self.lock:
			has = gid in self.wss
		return has

	def create (self, gid, *args):
		with self.lock:
			if self._nomore:
				return

		wss = WebSocketServer (gid, *args)
		with self.lock:
			self.wss [gid] = wss
		return wss

	def remove (self, gid):
		with self.lock:
			try:
				del self.wss [gid]
			except KeyError:
				pass

	def close (self):
		with self.lock:
			wss = list (self.wss.items ())
			self._nomore = True
		for k, s in wss:
			self.wasc.logger ('server', '...closing websockket %s' % k)
			s.close ()

	def cleanup (self):
		self.close ()


class WebSocketServer (specs.WebSocket1):
	def __init__ (self, gid, handler, request, apph, env, message_encoding = None):
		specs.WebSocket.__init__ (self, handler, request, message_encoding)
		self._closed = False
		self.gid = gid
		self.apph = apph
		self.env = env
		self.messages = []
		self.clients = {}
		self.client_id = None

	def add_client (self, ws):
		self.clients [ws.client_id] = ws
		ws.handle_message (1, skitai.WS_EVT_OPEN)

	def handle_client (self, client_id, event):
		if event == skitai.WS_EVT_CLOSE: # exit
			try: del self.clients [client_id]
			except KeyError: pass
			if not self.clients:
				return self.close ()

	def close (self):
		websocket_servers.remove (self.gid)
		self.clients = {}
		self._closed = True

	def send (self, msg, client_id = None, op_code = -1):
		msg, op_code = self.build_data (msg, op_code)
		if client_id:
			self.__sendto (client_id, msg, op_code)
		else:
			self.__sendall (msg, op_code)

	def __sendall (self, msg, op_code = -1):
		clients = list (self.clients.keys ())
		for client_id in clients:
			self.__sendto (client_id, msg, op_code)

	def __sendto (self, client_id, msg, op_code = -1):
		try:
			client = self.clients [client_id]
		except KeyError:
			client = None
		if client:
			client.send (msg, op_code)

websocket_servers = None

def start_websocket (wasc):
	global websocket_servers

	websocket_servers = WebSocketServers (wasc)





#---------------------------------------------------------------
# Skitai Unit Test Libaray
#---------------------------------------------------------------

import threading
import multiprocessing
from aquests.athreads import threadlib, trigger
import skitai
from skitai import was as the_was
from ..wastuff import triple_logger
from .. import wsgiappservice, cachefs
from ..handlers.websocket import servers as websocekts
from . import server
from ..handlers import vhost_handler
from ..handlers import proxy_handler
import inspect
import os
from ..corequest.dbi import task as dtask
from ..corequest.httpbase import task, cluster_manager
from aquests.client import socketpool
from aquests.dbapi import dbpool
from aquests import lifetime
from aquests.client import adns
from aquests.client.asynconnect import AsynConnect
from aquests.dbapi.dbconnect import DBConnect
from rs4 import asyncore
from .launcher  import launch
from skitai import PROTO_HTTP, PROTO_HTTPS, PROTO_WS, DB_PGSQL, DB_SQLITE3, DB_MONGODB, DB_REDIS, DB_SYN_PGSQL, DB_SYN_MONGODB, DB_SYN_REDIS
from ..wastuff import semaps
from ..corequest.pth import executors

def logger ():
    return triple_logger.Logger ("screen", None)    

#----------------------------------------------------------
def disable_threads ():    
    [the_was.queue.put (None) for each in range (the_was.numthreads)]
    the_was.queue = None
    the_was.threads = None
    the_was.numthreads = 0

def enable_threads (numthreads = 1):
    queue = threadlib.request_queue2 ()
    the_was.queue =  queue
    the_was.threads = threadlib.thread_pool (queue, numthreads, the_was.logger.get ("server"))
    the_was.numthreads = numthreads

#----------------------------------------------------------

def mount (point, target, pref = None):     
    global wasc
    
    for h in wasc.httpserver.handlers:
        if isinstance (h, vhost_handler.Handler):
            vh = h
            break
    
    if isinstance (target, str):
        target = "{} = {}".format (point, target)
    else:
        target = (point,) + target    
    vh.add_route ("default", target, pref)
    
def install_vhost_handler (apigateway = 0, apigateway_authenticate = 0):
    global wasc

    wasc.httpserver.handlers = []            
    static_max_ages = {"/img": 3600}    
    enable_apigateway = apigateway
    apigateway_authenticate = apigateway_authenticate 
    apigateway_realm = "Pytest"
    apigateway_secret_key = "secret-pytest"
    
    vh = wasc.add_handler (
        1, 
        vhost_handler.Handler, 
        wasc.clusters, 
        wasc.cachefs, 
        static_max_ages,
        enable_apigateway,
        apigateway_authenticate,
        apigateway_realm,
        apigateway_secret_key
    )    
    return vh

def install_proxy_handler ():
    global wasc
    
    wasc.httpserver.handlers = []
    h = wasc.add_handler (
        1, 
        proxy_handler.Handler, 
        wasc.clusters, 
        wasc.cachefs, 
        False
    )    
    return h

#----------------------------------------------------------

SAMPLE_DBPATH = '/tmp/example.sqlite'

def setup_was (wasc):
    def add_cluster (wasc, name, args):
        ctype, members, policy, ssl, max_conns = args
        wasc.add_cluster (ctype, name, members, ssl, policy, max_conns)
    
    # was and testutil was share objects
    try:
        wasc.register ("logger", logger ())
    except AttributeError: # class has been already setup
        return wasc
        
    wasc.register ("httpserver", server.Server (wasc.logger))
    wasc.register ("debug", False)
    
    wasc.register ("clusters",  {})
    wasc.register ("clusters_for_distcall",  {"__socketpool__": None, "__dbpool__": None})
    wasc.register ("workers", 1)
    wasc.register ("cachefs", cachefs.CacheFileSystem (None, 0, 0))    
    wasc.register ("executors", executors.Executors ())

    websocekts.start_websocket (wasc)
    wasc.register ("websockets", websocekts.websocket_servers)
    
    add_cluster (wasc, *skitai.alias ("@example", PROTO_HTTP, "www.example.com"))
    add_cluster (wasc, *skitai.alias ("@examples", PROTO_HTTPS, "www.example.com"))
    add_cluster (wasc, *skitai.alias ("@sqlite3", DB_SQLITE3, SAMPLE_DBPATH))
    add_cluster (wasc, *skitai.alias ("@postgresql", DB_PGSQL, "user:pass@127.0.0.1/mydb"))
    add_cluster (wasc, *skitai.alias ("@mongodb", DB_MONGODB, "127.0.0.1:27017/mydb"))
    add_cluster (wasc, *skitai.alias ("@redis", DB_REDIS, "127.0.0.1:6379"))
       
    return wasc

wasc = None
def activate (make_sync = True):
    from ..wsgiappservice import WAS
    from atila import was as atila_was
    
    class WAS (atila_was.WAS, WAS):
        numthreads = 1 
        _luwatcher = semaps.TestSemaps ()
    
    global wasc    
    if wasc is not None:
        return
    
    # convert async to sync
    if make_sync:
        cluster_manager.ClusterManager.use_syn_connection = True
        socketpool.SocketPool.use_syn_connection = True
        class_map = dbpool.DBPool.class_map
        class_map [DB_PGSQL] = class_map [DB_SYN_PGSQL]
        class_map [DB_REDIS] = class_map [DB_SYN_REDIS]
        class_map [DB_MONGODB] = class_map [DB_SYN_MONGODB]
    
    wasc = setup_was (WAS)
    skitai.start_was (wasc, enable_requests = True)   
    wasc._luwatcher.add (skitai.dconf ["models_keys"])
    lifetime.init (10.0, wasc.logger.get ("server"))    
    
    